#include <fors_data.h>
#include <fors_pfits.h>
#include <fors_dfs.h>

#include <cpl.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>

#define HARDCODED 0 /* create same values as MIDAS? */

#define DIE \
do { \
    fprintf(stderr, "Error at %s:%d\n", __func__, __LINE__); \
    if (errno) perror(NULL); \
    cpl_end(); \
    exit(EXIT_FAILURE); \
} while (0)

#define CHECK \
do { \
    if (cpl_error_get_code()) { \
        cpl_msg_error(cpl_func, "%s %s", cpl_error_get_where(), cpl_error_get_message()); \
        DIE; \
    } \
} while (0)

int main(int argc, char *argv[])
{
    cpl_init(CPL_INIT_DEFAULT);

    errno = 0;

    if (argc != 1) {
        fprintf(stderr,
                "Usage:\n    %s\n", argv[0]);
        DIE;
    }
    
    /* Numbers and code imported from MIDAS: fspil/pipe/proc/fors_zeropoint.prg */

    // FORS1 data
//    if (fors1upgraded == 0) {

    /* Old fors1 */
    double exti_old_u =  0.425;  // Extinction in the U band
    double exti_old_b =  0.245;  // Extinction in the B band
    double exti_old_v =  0.125;  // Extinction in the V band
    double exti_old_r =  0.081;  // Extinction in the R band
    double exti_old_i =  0.038;  // Extinction in the I band
  
    double colo_old_u =  0.067;  // Color term in the U band
    double colo_old_b = -0.078;  // Color term in the B band
    double colo_old_v =  0.035;  // Color term in the V band
    double colo_old_r =  0.042;  // Color term in the R band
    double colo_old_i = -0.038;  // Color term in the I band

    double exti_old_u_high =  0.000;  // Extinction in the u band
    double exti_old_b_high =  0.000;  // Extinction in the b band
    double exti_old_v_high =  0.000;  // Extinction in the v band
    double exti_old_g_high =  0.000;  // Extinction in the g band

    double colo_old_u_high =  0.000;  // Color term in the u band
    double colo_old_b_high =  0.000;  // Color term in the b band
    double colo_old_v_high =  0.000;  // Color term in the v band
    double colo_old_g_high =  0.000;  // Color term in the g band

    double zero_old_u = 99.99; // not available
    double zero_old_b = 99.99;
    double zero_old_v = 99.99;
    double zero_old_r = 99.99;
    double zero_old_i = 99.99;
    double zero_old_g = 99.99;

    struct {
        const char *instrument;
        struct {
            const char *base_filename;
            const char *chip_id;
            struct { 
                const char *filtername;
                enum filter filter;
                double exti, dexti;
                double colo, dcolo;
                double zero, dzero;
            } data[9];
        } chips[2];
    }
    dat[] = {
        /* Old FORS1 */
        {"FORS1", 
         {
             {"fors1_TK", "TK2048EB4-1",
              {{"U_BESS", FILTER_U, .exti = exti_old_u     , .dexti = 0, .colo = colo_old_u     , .dcolo = 0, .zero = zero_old_u, .dzero = 0},
               {"u_HIGH", FILTER_U, .exti = exti_old_u_high, .dexti = 0, .colo = colo_old_u_high, .dcolo = 0, .zero = zero_old_u, .dzero = 0},
               {"B_BESS", FILTER_B, .exti = exti_old_b     , .dexti = 0, .colo = colo_old_b     , .dcolo = 0, .zero = zero_old_b, .dzero = 0},
               {"b_HIGH", FILTER_B, .exti = exti_old_b_high, .dexti = 0, .colo = colo_old_b_high, .dcolo = 0, .zero = zero_old_b, .dzero = 0},
               {"g_HIGH", FILTER_G, .exti = exti_old_g_high, .dexti = 0, .colo = colo_old_g_high, .dcolo = 0, .zero = zero_old_g, .dzero = 0},
               {"V_BESS", FILTER_V, .exti = exti_old_v     , .dexti = 0, .colo = colo_old_v     , .dcolo = 0, .zero = zero_old_v, .dzero = 0},
               {"v_HIGH", FILTER_V, .exti = exti_old_v_high, .dexti = 0, .colo = colo_old_v_high, .dcolo = 0, .zero = zero_old_v, .dzero = 0},
               {"R_BESS", FILTER_R, .exti = exti_old_r     , .dexti = 0, .colo = colo_old_r     , .dcolo = 0, .zero = zero_old_r, .dzero = 0},
               {"I_BESS", FILTER_I, .exti = exti_old_i     , .dexti = 0, .colo = colo_old_i     , .dcolo = 0, .zero = zero_old_i, .dzero = 0}}
             },
             {NULL, NULL, /* Not used */
              {{"U_BESS", FILTER_U, .exti = exti_old_u     , .dexti = 0, .colo = colo_old_u     , .dcolo = 0, .zero = zero_old_u, .dzero = 0},
               {"u_HIGH", FILTER_U, .exti = exti_old_u_high, .dexti = 0, .colo = colo_old_u_high, .dcolo = 0, .zero = zero_old_u, .dzero = 0},
               {"B_BESS", FILTER_B, .exti = exti_old_b     , .dexti = 0, .colo = colo_old_b     , .dcolo = 0, .zero = zero_old_b, .dzero = 0},
               {"b_HIGH", FILTER_B, .exti = exti_old_b_high, .dexti = 0, .colo = colo_old_b_high, .dcolo = 0, .zero = zero_old_b, .dzero = 0},
               {"g_HIGH", FILTER_G, .exti = exti_old_g_high, .dexti = 0, .colo = colo_old_g_high, .dcolo = 0, .zero = zero_old_g, .dzero = 0},
               {"V_BESS", FILTER_V, .exti = exti_old_v     , .dexti = 0, .colo = colo_old_v     , .dcolo = 0, .zero = zero_old_v, .dzero = 0},
               {"v_HIGH", FILTER_V, .exti = exti_old_v_high, .dexti = 0, .colo = colo_old_v_high, .dcolo = 0, .zero = zero_old_v, .dzero = 0},
               {"R_BESS", FILTER_R, .exti = exti_old_r     , .dexti = 0, .colo = colo_old_r     , .dcolo = 0, .zero = zero_old_r, .dzero = 0},
               {"I_BESS", FILTER_I, .exti = exti_old_i     , .dexti = 0, .colo = colo_old_i     , .dcolo = 0, .zero = zero_old_i, .dzero = 0}}
             }
         }
        },
        /* New FORS1 */

#if HARDCODED
        /* Old values, hardcoded in MIDAS */

        {"FORS1", 
         {
             {"fors1_Norma", "Norma III",
              {{"U_BESS", FILTER_U, .exti = 0.456, .dexti = 0, .colo =  0.091, .dcolo = 0, .zero = 26.080, .dzero = 0.031},
               {"u_HIGH", FILTER_U, .exti = 0.440, .dexti = 0, .colo =  0.106, .dcolo = 0, .zero = 26.572, .dzero = 0.049},
               {"B_BESS", FILTER_B, .exti = 0.255, .dexti = 0, .colo = -0.107, .dcolo = 0, .zero = 28.163, .dzero = 0.023},
               {"b_HIGH", FILTER_B, .exti = 0.230, .dexti = 0, .colo =  0.069, .dcolo = 0, .zero = 28.539, .dzero = 0.031},
               {"g_HIGH", FILTER_G, .exti = 0.182, .dexti = 0, .colo =  0.031, .dcolo = 0, .zero = 28.867, .dzero = 0.032},
               {"V_BESS", FILTER_V, .exti = 0.132, .dexti = 0, .colo =  0.020, .dcolo = 0, .zero = 28.227, .dzero = 0.023},
               {"v_HIGH", FILTER_V, .exti = 0.115, .dexti = 0, .colo =  0.049, .dcolo = 0, .zero = 28.410, .dzero = 0.029},
               {"R_BESS", FILTER_R, .exti = 0.090, .dexti = 0, .colo =  0.000, .dcolo = 0, .zero = 28.024, .dzero = 0.025},
               {"I_BESS", FILTER_I, .exti = 0.051, .dexti = 0, .colo = -0.060, .dcolo = 0, .zero = 27.023, .dzero = 0.034}}
             },
             {"fors1_Marlene", "Marlene ",
              {{"U_BESS", FILTER_U, .exti = 0.456, .dexti = 0, .colo =  0.091, .dcolo = 0, .zero = 26.080, .dzero = 0.031},
               {"u_HIGH", FILTER_U, .exti = 0.440, .dexti = 0, .colo =  0.106, .dcolo = 0, .zero = 26.572, .dzero = 0.049},
               {"B_BESS", FILTER_B, .exti = 0.255, .dexti = 0, .colo = -0.107, .dcolo = 0, .zero = 28.163, .dzero = 0.023},
               {"b_HIGH", FILTER_B, .exti = 0.230, .dexti = 0, .colo =  0.069, .dcolo = 0, .zero = 28.539, .dzero = 0.031},
               {"g_HIGH", FILTER_G, .exti = 0.182, .dexti = 0, .colo =  0.031, .dcolo = 0, .zero = 28.867, .dzero = 0.032},
               {"V_BESS", FILTER_V, .exti = 0.132, .dexti = 0, .colo =  0.020, .dcolo = 0, .zero = 28.227, .dzero = 0.023},
               {"v_HIGH", FILTER_V, .exti = 0.115, .dexti = 0, .colo =  0.049, .dcolo = 0, .zero = 28.410, .dzero = 0.029},
               {"R_BESS", FILTER_R, .exti = 0.090, .dexti = 0, .colo =  0.000, .dcolo = 0, .zero = 28.024, .dzero = 0.025},
               {"I_BESS", FILTER_I, .exti = 0.051, .dexti = 0, .colo = -0.060, .dcolo = 0, .zero = 27.023, .dzero = 0.034}}
             },
         }
        },
        
        {"FORS2",
         {
             {"fors2_1453", "CCID20-14-5-3",
              {{"U_SPECIAL", FILTER_U, .exti = 0.443, .dexti = 0, .colo =  0.076, .dcolo = 0, .zero = 24.365, .dzero = 0.045},
               {"B_BESS"   , FILTER_B, .exti = 0.217, .dexti = 0, .colo = -0.033, .dcolo = 0, .zero = 27.740, .dzero = 0.032},
               {"V_BESS"   , FILTER_V, .exti = 0.127, .dexti = 0, .colo =  0.022, .dcolo = 0, .zero = 28.186, .dzero = 0.025},
               {"R_SPECIAL", FILTER_R, .exti = 0.098, .dexti = 0, .colo = -0.009, .dcolo = 0, .zero = 28.404, .dzero = 0.037},
               {"I_BESS"   , FILTER_I, .exti = 0.060, .dexti = 0, .colo = -0.017, .dcolo = 0, .zero = 27.773, .dzero = 0.034},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0}}
             },
             
             {"fors2_1456", "CCID20-14-5-6",
              {{"U_SPECIAL", FILTER_U, .exti = 0.443, .dexti = 0, .colo =  0.076, .dcolo = 0, .zero = 24.365, .dzero = 0.045},
               {"B_BESS"   , FILTER_B, .exti = 0.217, .dexti = 0, .colo = -0.033, .dcolo = 0, .zero = 27.740, .dzero = 0.032},
               {"V_BESS"   , FILTER_V, .exti = 0.127, .dexti = 0, .colo =  0.022, .dcolo = 0, .zero = 28.186, .dzero = 0.025},
               {"R_SPECIAL", FILTER_R, .exti = 0.098, .dexti = 0, .colo = -0.009, .dcolo = 0, .zero = 28.404, .dzero = 0.037},
               {"I_BESS"   , FILTER_I, .exti = 0.060, .dexti = 0, .colo = -0.017, .dcolo = 0, .zero = 27.773, .dzero = 0.034},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0}}
             }
         }
        }

#else
        /* Numbers from P79 as provided by DFO */

        {"FORS1", 
         {
             {"fors1_Norma", "Norma III",
              {{"U_BESS", FILTER_U, .exti = 0.447, .dexti = 0.018, .colo =  0.043, .dcolo = 0.008, .zero = 26.080, .dzero = 0.031},
               {"u_HIGH", FILTER_U, .exti = 0.437, .dexti = 0.033, .colo =  0.066, .dcolo = 0.012, .zero = 26.572, .dzero = 0.049},
               {"B_BESS", FILTER_B, .exti = 0.241, .dexti = 0.013, .colo = -0.128, .dcolo = 0.008, .zero = 28.163, .dzero = 0.023},
               {"b_HIGH", FILTER_B, .exti = 0.222, .dexti = 0.020, .colo =  0.048, .dcolo = 0.009, .zero = 28.539, .dzero = 0.031},
               {"g_HIGH", FILTER_G, .exti = 0.209, .dexti = 0.022, .colo = -0.096, .dcolo = 0.008, .zero = 28.867, .dzero = 0.032},
               {"V_BESS", FILTER_V, .exti = 0.112, .dexti = 0.012, .colo = -0.015, .dcolo = 0.007, .zero = 28.227, .dzero = 0.023},
               {"v_HIGH", FILTER_V, .exti = 0.139, .dexti = 0.018, .colo =  0.019, .dcolo = 0.008, .zero = 28.410, .dzero = 0.029},
               {"R_BESS", FILTER_R, .exti = 0.087, .dexti = 0.014, .colo = -0.097, .dcolo = 0.015, .zero = 28.024, .dzero = 0.025},
               {"I_BESS", FILTER_I, .exti = 0.027, .dexti = 0.021, .colo = -0.101, .dcolo = 0.011, .zero = 27.023, .dzero = 0.034}}
             },
             {"fors1_Marlene", "Marlene ",
              {{"U_BESS", FILTER_U, .exti = 0.470, .dexti = 0.024, .colo =  0.039, .dcolo = 0.013, .zero = 26.069, .dzero = 0.041},
               {"u_HIGH", FILTER_U, .exti = 0.445, .dexti = 0.051, .colo =  0.063, .dcolo = 0.018, .zero = 26.548, .dzero = 0.070},
               {"B_BESS", FILTER_B, .exti = 0.316, .dexti = 0.019, .colo = -0.124, .dcolo = 0.012, .zero = 28.221, .dzero = 0.032},
               {"b_HIGH", FILTER_B, .exti = 0.282, .dexti = 0.036, .colo =  0.059, .dcolo = 0.016, .zero = 28.563, .dzero = 0.052},
               {"g_HIGH", FILTER_G, .exti = 0.287, .dexti = 0.063, .colo = -0.106, .dcolo = 0.019, .zero = 28.938, .dzero = 0.082},
               {"V_BESS", FILTER_V, .exti = 0.158, .dexti = 0.018, .colo = -0.011, .dcolo = 0.012, .zero = 28.247, .dzero = 0.033},
               {"v_HIGH", FILTER_V, .exti = 0.155, .dexti = 0.030, .colo =  0.019, .dcolo = 0.014, .zero = 28.388, .dzero = 0.046},
               {"R_BESS", FILTER_R, .exti = 0.111, .dexti = 0.018, .colo = -0.047, .dcolo = 0.022, .zero = 28.009, .dzero = 0.033},
               {"I_BESS", FILTER_I, .exti = 0.037, .dexti = 0.019, .colo = -0.100, .dcolo = 0.012, .zero = 27.038, .dzero = 0.035}}
             },
         }
        },
        
        {"FORS2",
         {
             {"fors2_1453", "CCID20-14-5-3",
              {{"U_SPECIAL", FILTER_U, .exti = 0.531, .dexti = 0.025, .colo =  0.069, .dcolo = 0.009, .zero = 24.365, .dzero = 0.045},
               {"B_BESS"   , FILTER_B, .exti = 0.236, .dexti = 0.015, .colo = -0.026, .dcolo = 0.009, .zero = 27.740, .dzero = 0.032},
               {"V_BESS"   , FILTER_V, .exti = 0.117, .dexti = 0.012, .colo = -0.002, .dcolo = 0.006, .zero = 28.186, .dzero = 0.025},
               {"R_SPECIAL", FILTER_R, .exti = 0.067, .dexti = 0.020, .colo = -0.039, .dcolo = 0.016, .zero = 28.404, .dzero = 0.037},
               {"I_BESS"   , FILTER_I, .exti = 0.049, .dexti = 0.016, .colo = -0.044, .dcolo = 0.007, .zero = 27.773, .dzero = 0.034},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0}}
             },
             
             {"fors2_1456", "CCID20-14-5-6",
              {{"U_SPECIAL", FILTER_U, .exti = 0.526, .dexti = 0.042, .colo =  0.067, .dcolo = 0.014, .zero = 24.476, .dzero = 0.066},
               {"B_BESS"   , FILTER_B, .exti = 0.272, .dexti = 0.023, .colo = -0.062, .dcolo = 0.011, .zero = 27.837, .dzero = 0.040},
               {"V_BESS"   , FILTER_V, .exti = 0.123, .dexti = 0.014, .colo = -0.008, .dcolo = 0.006, .zero = 28.207, .dzero = 0.028},
               {"R_SPECIAL", FILTER_R, .exti = 0.065, .dexti = 0.022, .colo = -0.036, .dcolo = 0.013, .zero = 28.400, .dzero = 0.037},
               {"I_BESS"   , FILTER_I, .exti = 0.023, .dexti = 0.013, .colo = -0.080, .dcolo = 0.008, .zero = 27.777, .dzero = 0.039},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0},
               {NULL , FILTER_I, 0, 0, 0, 0, 0, 0}}
             }
         }
        }
#endif
    };
    
    CHECK;

#if 1

    /* For each instrument */
    unsigned i;
    for(i = 0; i < sizeof(dat) / sizeof(*dat); i++) {
        
        /* For each chip */
        unsigned chips;
        for(chips = 0; 
            chips < sizeof(dat[i].chips) / sizeof(*dat[i].chips); 
            chips++) {
            
            if (dat[i].chips[chips].chip_id == NULL) break;
            
            cpl_table *t = cpl_table_new(9999);
            int size = 0;
        
            cpl_propertylist *header = cpl_propertylist_new();
        
            cpl_propertylist_update_string(header, FORS_PFITS_CHIP_ID, dat[i].chips[chips].chip_id);
            cpl_propertylist_update_string(header, "ESO PRO CATG", PHOT_TABLE);
            cpl_propertylist_update_string(header, FORS_PFITS_INSTRUME, dat[i].instrument);


            //cpl_table_new_column(t, "Instrument" , CPL_TYPE_STRING);
            //cpl_table_new_column(t, "Chip" , CPL_TYPE_STRING);
            cpl_table_new_column(t, FORS_DATA_PHOT_FILTER , CPL_TYPE_STRING);
            cpl_table_new_column(t, FORS_DATA_PHOT_EXTCOEFF , CPL_TYPE_DOUBLE);
            cpl_table_new_column(t, FORS_DATA_PHOT_DEXTCOEFF , CPL_TYPE_DOUBLE);
            cpl_table_new_column(t, FORS_DATA_PHOT_COLORTERM, CPL_TYPE_DOUBLE);
            cpl_table_new_column(t, FORS_DATA_PHOT_DCOLORTERM, CPL_TYPE_DOUBLE);
            cpl_table_new_column(t, FORS_DATA_PHOT_ZEROPOINT, CPL_TYPE_DOUBLE);
            cpl_table_new_column(t, FORS_DATA_PHOT_DZEROPOINT, CPL_TYPE_DOUBLE);
        
      
            /* For each filter */
            unsigned j;
            for(j = 0; 
                j < sizeof(dat[i].chips[chips].data) / sizeof(*dat[i].chips[chips].data); 
                j++)
                {            
                    if (dat[i].chips[chips].data[j].filtername == NULL) break;
                           
                    //cpl_table_set_string(t, "Instrument", size, dat[i].instrument);
                    //cpl_table_set_string(t, "Chip", size, dat[i].chips[chips].chip_id);
                    cpl_table_set_string(t, FORS_DATA_PHOT_FILTER, size, dat[i].chips[chips].data[j].filtername);
                    cpl_table_set_double(t, FORS_DATA_PHOT_EXTCOEFF , size, 
                                         dat[i].chips[chips].data[j].exti);
                    cpl_table_set_double(t, FORS_DATA_PHOT_COLORTERM, size, 
                                         dat[i].chips[chips].data[j].colo);
                    cpl_table_set_double(t, FORS_DATA_PHOT_ZEROPOINT, size,
                                         dat[i].chips[chips].data[j].zero);
                    cpl_table_set_double(t, FORS_DATA_PHOT_DEXTCOEFF , size, 
                                         dat[i].chips[chips].data[j].dexti);
                    cpl_table_set_double(t, FORS_DATA_PHOT_DCOLORTERM, size, 
                                         dat[i].chips[chips].data[j].dcolo);
                    cpl_table_set_double(t, FORS_DATA_PHOT_DZEROPOINT, size,
                                         dat[i].chips[chips].data[j].dzero);
                    
                    size++;
                }
            
            cpl_table_set_size(t, size);
            //cpl_table_dump(t, 0, size, stdout);

#if HARDCODED
            const char *full_filename = cpl_sprintf("%s_phot_old.fits", 
                                                    dat[i].chips[chips].base_filename);
#else
            const char *full_filename = cpl_sprintf("%s_phot.fits", 
                                                    dat[i].chips[chips].base_filename);
#endif
            
            cpl_table_save(t, header, NULL, full_filename, CPL_IO_DEFAULT);
            
            CHECK;  
            fprintf(stdout, "Created %s\n", full_filename);

            cpl_table_delete(t); t = NULL;
            cpl_propertylist_delete(header); header = NULL;
            cpl_free((void *)full_filename);
        }  
    }

#else 
    /* One big table */

    cpl_table *t = cpl_table_new(9999);
    int size = 0;

    cpl_propertylist *header = cpl_propertylist_new();
                
    cpl_table_new_column(t, "Instrument" , CPL_TYPE_STRING);
    cpl_table_new_column(t, FORS_DATA_PHOT_FILTER , CPL_TYPE_STRING);
    cpl_table_new_column(t, "Chip" , CPL_TYPE_STRING);
    cpl_table_new_column(t, FORS_DATA_PHOT_EXTCOEFF , CPL_TYPE_DOUBLE);
    cpl_table_new_column(t, FORS_DATA_PHOT_COLORTERM, CPL_TYPE_DOUBLE);
    cpl_table_new_column(t, FORS_DATA_PHOT_ZEROPOINT, CPL_TYPE_DOUBLE);
    
    unsigned i;
    for(i = 0; i < sizeof(dat) / sizeof(*dat); i++) {
        
        unsigned j;
        for(j = 0; 
            j < sizeof(dat[i].chips[chips].data) / sizeof(*dat[i].chips[chips].data); 
            j++) {
            
            if (dat[i].chips[chips].data[j].filtername == NULL) break;
            
            
            unsigned chips;
            for(chips = 0; 
                chips < sizeof(dat[i].chip_id) / sizeof(*dat[i].chip_id); 
                chips++) {
                
                if (dat[i].chip_id[chips] == NULL) break;
                
                cpl_table_set_string(t, "Instrument", size, dat[i].instrument);
                cpl_table_set_string(t, "Chip", size, dat[i].chip_id[chips]);
                cpl_table_set_string(t, FORS_DATA_PHOT_FILTER, size, dat[i].chips[chips].data[j].filtername);
                cpl_table_set_double(t, FORS_DATA_PHOT_EXTCOEFF , size, 
                                     dat[i].chips[chips].data[j].exti);
                cpl_table_set_double(t, FORS_DATA_PHOT_COLORTERM, size, 
                                     dat[i].chips[chips].data[j].colo);
                cpl_table_set_double(t, FORS_DATA_PHOT_ZEROPOINT, size,
                                     dat[i].chips[chips].data[j].zero);
                
                size++;
                //cpl_table_dump(t, 0, size, stdout);
            }
        }
    }
    
    cpl_table_set_size(t, size);

    cpl_table_save(t, NULL, NULL, filename, CPL_IO_DEFAULT);

    CHECK;  
    fprintf(stdout, "Created %s\n", filename);


#endif







#if 0





    t = cpl_table_new(FORS_NUM_FILTER);

    /* Here the order must match fors_data.h: enum filter */
    double *exti[FORS_NUM_FILTER] = {&exti_u,
                                     &exti_b,
                                     NULL,   // g
                                     &exti_v,
                                     &exti_r,
                                     &exti_i,
                                     NULL};   // z

    double *colo[FORS_NUM_FILTER] = {&colo_u,
                                     &colo_b,
                                     NULL,   // g
                                     &colo_v,
                                     &colo_r,
                                     &colo_i,
                                     NULL};   // z

    double *zero[FORS_NUM_FILTER] = {&zero_u,
                                     &zero_b,
                                     NULL,
                                     &zero_v,
                                     &zero_r,
                                     &zero_i,
                                     NULL};

    cpl_table_new_column(t, FORS_DATA_PHOT_FILTER   , CPL_TYPE_STRING);
    cpl_table_new_column(t, FORS_DATA_PHOT_EXTCOEFF , CPL_TYPE_DOUBLE);
    cpl_table_new_column(t, FORS_DATA_PHOT_COLORTERM, CPL_TYPE_DOUBLE);
    cpl_table_new_column(t, FORS_DATA_PHOT_ZEROPOINT, CPL_TYPE_DOUBLE);

    int row = 0;
    {
        int i;
        for(i = 0; i < cpl_table_get_nrow(t); i++) {
            if (exti[i] != NULL && colo[i] != NULL && zero[i] != NULL) {
                cpl_table_set_string(t, FORS_DATA_PHOT_FILTER  , row, FORS_DATA_STD_MAG[i]);
                cpl_table_set_double(t, FORS_DATA_PHOT_EXTCOEFF, row, *(exti[i]));
                cpl_table_set_double(t, FORS_DATA_PHOT_COLORTERM, row, *(colo[i]));
                cpl_table_set_double(t, FORS_DATA_PHOT_ZEROPOINT, row, *(zero[i]));
                row++;
            }
        }
    }

    CHECK;
   
    cpl_table_set_size(t, row);

    cpl_table_save(t, NULL, NULL, filename, CPL_IO_DEFAULT);

    CHECK;
    
    fprintf(stdout, "Created %s\n", filename);
    cpl_table_delete(t);

#endif

    return 0;
}
