#include<cpl.h>

#include<stdio.h>
#include<string.h>
#include<errno.h>

/*
   This program creates a standard star FITS table from Stetson's list of ASCII data


  Usage

  1. Get all Stetson  *.pho and *.pos files
  2. Filter out headers, append to single file:
        $ cat *.pho | grep -v "vary?"   > stetson.pho
        $ cat *.pos | grep -v Reference > stetson.pos
  3. Convert to FITS
        $ create_stetson_fits stetson.pho stetson.pos stetson.fits
     This program will verify that the object names in the 2 ASCII 
     files match row by row .

*/

/*
  This research used the facilities of the Canadian Astronomy Data Centre operated
  by the National Research Council of Canada with the support of the Canadian Space Agency.
*/

#define DIE \
do { \
    fprintf(stderr, "Error at %s:%d\n", __func__, __LINE__); \
    if (errno) perror(NULL); \
    cpl_end(); \
    exit(EXIT_FAILURE); \
} while (0)

#define CHECK \
do { \
    if (cpl_error_get_code()) { \
        cpl_msg_error(cpl_func, "%s", cpl_error_get_where()); \
        DIE; \
    } \
} while (0)

int main(int argc, char *argv[])
{
    cpl_table *table;
    int size = 0;
    const int maxsize = 100*1000;

    cpl_init(CPL_INIT_DEFAULT);

    if (argc != 4) {
        fprintf(stderr,
                "Usage:\n    %s  [stetson.pho] [stetson.pos]  [stetson.fits]\n", argv[0]);
        DIE;
    }
    
    const char *filename_pho = argv[1];
    const char *filename_pos = argv[2];
    const char *filename_out = argv[3];

    table = cpl_table_new(maxsize);
    CHECK;
    
    struct {
        const char *colname;
        cpl_type type;
        const char *unit;
    } data[] = {

        {"OBJECT", CPL_TYPE_STRING, NULL},

        {"RA" , CPL_TYPE_DOUBLE, "degree"},
        {"DEC", CPL_TYPE_DOUBLE, "degree"},

        /* estimated intrinsic variability in magnitude RMS */
        {"VAR", CPL_TYPE_FLOAT, "mag rms"},

        /*  B:     magnitude */
        /*  ERR_B: magnitude error */
        /*  N_B:   number of observations obtained on photometric occasions */
        /*  n_B:   total number of observations */

        {"B"    , CPL_TYPE_FLOAT, "mag"},
        {"ERR_B", CPL_TYPE_FLOAT, "mag"},
        {"N_B"  , CPL_TYPE_INT, NULL},
        {"n_B"  , CPL_TYPE_INT, NULL},

        {"V"    , CPL_TYPE_FLOAT, "mag"},
        {"ERR_V", CPL_TYPE_FLOAT, "mag"},
        {"N_V"  , CPL_TYPE_INT, NULL},
        {"n_V"  , CPL_TYPE_INT, NULL},

        {"R"    , CPL_TYPE_FLOAT, "mag"},
        {"ERR_R", CPL_TYPE_FLOAT, "mag"},
        {"N_R"  , CPL_TYPE_INT, NULL},
        {"n_R"  , CPL_TYPE_INT, NULL},

        {"I"    , CPL_TYPE_FLOAT, "mag"},
        {"ERR_I", CPL_TYPE_FLOAT, "mag"},
        {"N_I"  , CPL_TYPE_INT, NULL},
        {"n_I"  , CPL_TYPE_INT, NULL}
        
    };
    
    {
        unsigned i;
        for (i = 0; i < sizeof(data) / sizeof(*data); i++) {
            cpl_table_new_column     (table, data[i].colname, data[i].type);
            if (data[i].unit != NULL) {
                cpl_table_set_column_unit(table, data[i].colname, data[i].unit);
            }
        }
    }
    CHECK;

    {
        char name_pho[1000];
        char name_pos[1000];
        float mag[5];
        float err[5];
        int N[5];
        int n[5];
        float var;
        float dummy;
        int items_assigned;

        FILE *file_pho = fopen(filename_pho, "r");
        FILE *file_pos = fopen(filename_pos, "r");

        if (file_pho == NULL) DIE;
        if (file_pos == NULL) DIE;
        
        size = 0;
        while((items_assigned = fscanf(file_pho, 
                                       "%s"
                                       "%f %f %d %d "
                                       "%f %f %d %d "
                                       "%f %f %d %d "
                                       "%f %f %d %d "
                                       "%f",
                                       name_pho, 
                                       &mag[0], &err[0], &N[0], &n[0],
                                       &mag[1], &err[1], &N[1], &n[1],
                                       &mag[2], &err[2], &N[2], &n[2],
                                       &mag[3], &err[3], &N[3], &n[3], &var))
              != EOF) {
            double ra, dec;
            int i;

            printf("%d: %s ", items_assigned, name_pho);
            printf("%f %f %d %d ", mag[0], err[0], N[0], n[0]);
            printf("%f %f %d %d ", mag[1], err[1], N[1], n[1]);
            printf("%f %f %d %d ", mag[2], err[2], N[2], n[2]);
            printf("%f %f %d %d ", mag[3], err[3], N[3], n[3]);
            printf("%f ", var);
                
            if (items_assigned != 18) DIE;


            /* Read RA, DEC */
            if (fscanf(file_pos,
                       "%lf %lf "
                       "%f %f %f "
                       "%f %f %f "
                       "%f %f %f %f "
                       "%s",
                       &ra, &dec,
                       &dummy, &dummy, &dummy, //hexagesimal RA
                       &dummy, &dummy, &dummy, //hexagesimal DEC
                       &dummy, &dummy, &dummy, &dummy,  //offset in arcsecs, pixels
                       name_pos) != 13) DIE;
                       
            printf("%s (%f, %f)", name_pos, ra, dec);
            printf("\n");

            if (strcmp(name_pho, name_pos) != 0) {
                fprintf(stderr, "Mismatching OBJECT name: %s (pho), %s (pos)\n",
                        name_pho, name_pos);
                DIE;                
            }
            
            cpl_table_set_string(table, data[0].colname, size, name_pho);
            cpl_table_set_double(table, data[1].colname, size, ra);
            cpl_table_set_double(table, data[2].colname, size, dec);
            cpl_table_set_float (table, data[3].colname, size, var);
            
            for (i = 0; i < 4; i++) {
                /* zero or two of N and n must be zero */
                if (N[i] == 0 && n[i] != 0) DIE;
                if (N[i] != 0 && n[i] == 0) DIE;
                
                if (N[i] != 0) {
                    cpl_table_set_float (table, data[i*4+4].colname, size, mag[i]);
                    cpl_table_set_float (table, data[i*4+5].colname, size, err[i]);
                    cpl_table_set_int   (table, data[i*4+6].colname, size,   N[i]);
                    cpl_table_set_int   (table, data[i*4+7].colname, size,   n[i]);
                } else {
                    cpl_table_set_invalid(table, data[i*4+4].colname, size);
                    cpl_table_set_invalid(table, data[i*4+5].colname, size);
                    cpl_table_set_invalid(table, data[i*4+6].colname, size);
                    cpl_table_set_invalid(table, data[i*4+7].colname, size);
                }
            }
            size++;
        }
        fclose(file_pho);

//        fclose(file_pos);
    }


    CHECK;

    cpl_table_set_size(table, size);
   
    cpl_table_save(table, NULL, NULL, filename_out, CPL_IO_DEFAULT);

    CHECK;
    
    fprintf(stdout, "Created %s\n", filename_out);

    return 0;
}
