/* $Id: fors_detlin.c,v 1.10 2010/09/14 07:38:16 cizzo Exp $
 *
 * This file is part of the FORS Data Reduction Pipeline
 * Copyright (C) 2002-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: cizzo $
 * $Date: 2010/09/14 07:38:16 $
 * $Revision: 1.10 $
 * $Name: fors-4_8_6 $
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------
                             Includes
 ----------------------------------------------------------------------------*/

#include <cpl.h>

#include "irplib_detmon.h"
#include "irplib_detmon_lg.h"
#include <fors_dfs.h>

/* Define here the DO.CATG keywords */
#define DETMON_OPT_LG_ON_RAW                    "DETECTOR_CHECK_FLAT"
#define DETMON_OPT_LG_OFF_RAW                   "DETECTOR_CHECK_BIAS"

#define RECIPE_NAME "fors_detlin"

#define OPT FALSE

/*----------------------------------------------------------------------------
                                Functions prototypes
 ----------------------------------------------------------------------------*/

static cpl_error_code
fors_detlin_fill_parlist(cpl_parameterlist *);

CPL_RECIPE_DEFINE(fors_detlin, FORS_BINARY_VERSION,
                  fors_detlin_fill_parlist(recipe->parameters),
                  "Lander de Bilbao", "lbilbao@eso.org", "2008", 
                  "Linearity/Gain recipe for the optical domain",
                  irplib_detmon_lg_get_description("fors_detlin", 
                                                   PACKAGE_TARNAME,
                                                   DETMON_OPT_LG_ON_RAW,
                                                   DETMON_OPT_LG_OFF_RAW));

/*---------------------------------------------------------------------------*/
/*
  @brief    Interpret the command line options and execute the data processing
  @param    frameset    the frames list
  @param    parlist     the parameters list
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int fors_detlin(cpl_frameset            *frameset,
                       const cpl_parameterlist *parlist)
{
    cpl_propertylist * lintbl = 
        irplib_detmon_fill_prolist("DET_LIN_INFO", NULL, NULL, CPL_FALSE);

    cpl_propertylist * gaintbl = 
        irplib_detmon_fill_prolist("GAIN_INFO",    NULL, NULL, CPL_FALSE);

    cpl_propertylist * coeffscube = 
        irplib_detmon_fill_prolist("COEFFS_CUBE",  NULL, NULL, CPL_FALSE);

    cpl_propertylist * bpm = 
        irplib_detmon_fill_prolist("BP_MAP_NL",    NULL, NULL, CPL_FALSE);

    cpl_propertylist * corr = 
        irplib_detmon_fill_prolist("AUTOCORR",     NULL, NULL, CPL_FALSE);

    cpl_propertylist * diff_flat = 
        irplib_detmon_fill_prolist("DIFF_FLAT",    NULL, NULL, CPL_FALSE);

    const cpl_error_code error = irplib_detmon_lg(frameset,
                                                  parlist,
                                                  DETMON_OPT_LG_ON_RAW,
                                                  DETMON_OPT_LG_OFF_RAW,
                                                  RECIPE_NAME,
                                                  PACKAGE_TARNAME,
                                                  REGEXP,
                                                  lintbl, gaintbl, coeffscube,
                                                  bpm, corr, diff_flat,
                                                  PACKAGE "/" PACKAGE_VERSION,
                                                  NULL, NULL, OPT);

    cpl_propertylist_delete(lintbl);
    cpl_propertylist_delete(gaintbl);
    cpl_propertylist_delete(coeffscube);
    cpl_propertylist_delete(bpm);
    cpl_propertylist_delete(corr);
    cpl_propertylist_delete(diff_flat);

    {
        /*
         * This part is added until ticket DFS05711 is fixed.
         */

        cpl_image *chipmap;
        cpl_image *badpixmap;
        int        llx = dfs_get_parameter_int(parlist, 
                         PACKAGE_TARNAME "." RECIPE_NAME ".llx", NULL);
        int        lly = dfs_get_parameter_int(parlist, 
                         PACKAGE_TARNAME "." RECIPE_NAME ".lly", NULL);

        badpixmap = dfs_load_image(frameset, "BP_MAP_NL", CPL_TYPE_FLOAT, 0, 1);
        chipmap = cpl_image_new(2048, 1024, CPL_TYPE_FLOAT);
        cpl_image_copy(chipmap, badpixmap, llx, lly - 5);
        cpl_image_delete(badpixmap);
        dfs_save_image(frameset, chipmap, "BP_MAP_NL_FULL", NULL, parlist, 
                       RECIPE_NAME, PACKAGE "/" PACKAGE_VERSION);
        cpl_image_delete(chipmap);
    }

    /* Propagate the error, if any */
    cpl_ensure_code(!error, error);

    return CPL_ERROR_NONE;
}

static cpl_error_code
fors_detlin_fill_parlist(cpl_parameterlist * parlist)
{
    cpl_error_code error;
    cpl_parameter * p;

    error = irplib_detmon_lg_fill_parlist_opt_default(parlist,
                                              "fors_detlin", PACKAGE_TARNAME);

    cpl_ensure_code(!error, error);

 
    p = cpl_parameterlist_find(parlist, 
                               PACKAGE_TARNAME "." RECIPE_NAME ".tolerance");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_double(p, 0.1);
    cpl_ensure_code(!error, error);

    p = cpl_parameterlist_find(parlist,
                               PACKAGE_TARNAME "." RECIPE_NAME ".pix2pix");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_bool(p, 1);
    cpl_ensure_code(!error, error);

    p = cpl_parameterlist_find(parlist,
                               PACKAGE_TARNAME "." RECIPE_NAME ".order");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_int(p, 2);
    cpl_ensure_code(!error, error);

    p = cpl_parameterlist_find(parlist,
                               PACKAGE_TARNAME "." RECIPE_NAME ".kappa");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_double(p, 4.0);
    cpl_ensure_code(!error, error);

/**/
    p = cpl_parameterlist_find(parlist,
                               PACKAGE_TARNAME "." RECIPE_NAME ".llx");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_int(p, 1);
    cpl_ensure_code(!error, error);

    p = cpl_parameterlist_find(parlist,
                               PACKAGE_TARNAME "." RECIPE_NAME ".urx");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_int(p, 2048);
    cpl_ensure_code(!error, error);
/**/

    p = cpl_parameterlist_find(parlist,
                               PACKAGE_TARNAME "." RECIPE_NAME ".lly");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_int(p, 6);
    cpl_ensure_code(!error, error);

    p = cpl_parameterlist_find(parlist,
                               PACKAGE_TARNAME "." RECIPE_NAME ".ury");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_int(p, 1029);
    cpl_ensure_code(!error, error);

/*
    p = cpl_parameterlist_find(parlist,
                               PACKAGE_TARNAME "." RECIPE_NAME ".method");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_string(p, "MED");
    cpl_ensure_code(!error, error);
*/

    return error;

}
