/* $Id: fors_flat.c,v 1.7 2010/09/14 07:38:16 cizzo Exp $
 *
 * This file is part of the FORS Data Reduction Pipeline
 * Copyright (C) 2002-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: cizzo $
 * $Date: 2010/09/14 07:38:16 $
 * $Revision: 1.7 $
 * $Name: fors-4_8_6 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <math.h>
#include <cpl.h>
#include <moses.h>
#include <fors_dfs.h>

static int fors_flat_create(cpl_plugin *);
static int fors_flat_exec(cpl_plugin *);
static int fors_flat_destroy(cpl_plugin *);
static int fors_flat(cpl_parameterlist *, cpl_frameset *);

static char fors_flat_description[] =
"This recipe is used to subtract the master bias, produced by the recipe\n"
"fors_bias, from a set of raw flat field frames. The input raw frames are\n"
"summed, the master bias frame is rescaled accordingly, and subtracted\n"
"from the result. The overscan regions, if present, are used to compensate\n"
"for variations of the bias level between master bias and input raw frames.\n"
"The overscan regions are then trimmed from the result.\n"
"In the table below the MXU acronym can be alternatively read as MOS and\n"
"LSS.\n\n"
"Input files:\n\n"
"  DO category:               Type:       Explanation:         Required:\n"
"  SCREEN_FLAT_MXU            Raw         Raw data frame          Y\n"
"  MASTER_BIAS                Calib       Master bias frame       Y\n\n"
"Output files:\n\n"
"  DO category:               Data type:  Explanation:\n"
"  MASTER_SCREEN_FLAT_MXU     FITS image  Bias subtracted sum frame\n\n";

#define fors_flat_exit(message)               \
{                                             \
if (message) cpl_msg_error(recipe, message);  \
cpl_image_delete(flat);                       \
cpl_image_delete(master_flat);                \
cpl_image_delete(master_bias);                \
cpl_propertylist_delete(header);              \
cpl_table_delete(overscans);                  \
cpl_msg_indent_less();                        \
return -1;                                    \
}

#define fors_flat_exit_memcheck(message)        \
{                                               \
if (message) cpl_msg_info(recipe, message);     \
printf("free flat (%p)\n", flat);               \
cpl_image_delete(flat);                         \
printf("free master_flat (%p)\n", master_flat); \
cpl_image_delete(master_flat);                  \
printf("free master_bias (%p)\n", master_bias); \
cpl_image_delete(master_bias);                  \
printf("free header (%p)\n", header);           \
cpl_propertylist_delete(header);                \
printf("free overscans (%p)\n", overscans);     \
cpl_table_delete(overscans);                    \
cpl_msg_indent_less();                          \
return 0;                                       \
}


/**
 * @brief    Build the list of available plugins, for this module. 
 *
 * @param    list    The plugin list
 *
 * @return   0 if everything is ok, -1 otherwise
 *
 * Create the recipe instance and make it available to the application 
 * using the interface. This function is exported.
 */

int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    FORS_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "fors_flat",
                    "Sum input flat field frames and remove bias",
                    fors_flat_description,
                    "Carlo Izzo",
                    PACKAGE_BUGREPORT,
    "This file is currently part of the FORS Instrument Pipeline\n"
    "Copyright (C) 2002-2010 European Southern Observatory\n\n"
    "This program is free software; you can redistribute it and/or modify\n"
    "it under the terms of the GNU General Public License as published by\n"
    "the Free Software Foundation; either version 2 of the License, or\n"
    "(at your option) any later version.\n\n"
    "This program is distributed in the hope that it will be useful,\n"
    "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n"
    "GNU General Public License for more details.\n\n"
    "You should have received a copy of the GNU General Public License\n"
    "along with this program; if not, write to the Free Software Foundation,\n"
    "Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA\n",
                    fors_flat_create,
                    fors_flat_exec,
                    fors_flat_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}


/**
 * @brief    Setup the recipe options    
 *
 * @param    plugin  The plugin
 *
 * @return   0 if everything is ok
 *
 * Defining the command-line/configuration parameters for the recipe.
 */

static int fors_flat_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;
/* Uncomment in case parameters are defined
    cpl_parameter *p;
*/

    /* 
     * Check that the plugin is part of a valid recipe 
     */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE) 
        recipe = (cpl_recipe *)plugin;
    else 
        return -1;

    /* 
     * Create the (empty) parameters list in the cpl_recipe object 
     */

    recipe->parameters = cpl_parameterlist_new(); 

    return 0;
}


/**
 * @brief    Execute the plugin instance given by the interface
 *
 * @param    plugin  the plugin
 *
 * @return   0 if everything is ok
 */

static int fors_flat_exec(cpl_plugin *plugin)
{
    cpl_recipe *recipe;
    
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE) 
        recipe = (cpl_recipe *)plugin;
    else 
        return -1;

    return fors_flat(recipe->parameters, recipe->frames);
}


/**
 * @brief    Destroy what has been created by the 'create' function
 *
 * @param    plugin  The plugin
 *
 * @return   0 if everything is ok
 */

static int fors_flat_destroy(cpl_plugin *plugin)
{
    cpl_recipe *recipe;
    
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE) 
        recipe = (cpl_recipe *)plugin;
    else 
        return -1;

    cpl_parameterlist_delete(recipe->parameters); 

    return 0;
}


/**
 * @brief    Interpret the command line options and execute the data processing
 *
 * @param    parlist     The parameters list
 * @param    frameset    The set-of-frames
 *
 * @return   0 if everything is ok
 */

static int fors_flat(cpl_parameterlist *parlist, cpl_frameset *frameset)
{

    const char *recipe = "fors_flat";


    /*
     * CPL objects
     */

    cpl_image        *flat        = NULL;
    cpl_image        *master_flat = NULL;
    cpl_image        *master_bias = NULL;
    cpl_image        *multi_bias  = NULL;
    cpl_image        *dummy       = NULL;
    cpl_table        *overscans   = NULL;
    cpl_propertylist *header      = NULL;

    /*
     * Auxiliary variables
     */

    char        version[80];
    const char *master_bias_tag = "MASTER_BIAS";
    const char *flat_tag;
    const char *pro_tag;
    char       *instrume = NULL;
    int         flat_mxu;
    int         flat_mos;
    int         flat_lss;
    int         nbias, nflat;
    int         i;


    cpl_msg_set_indentation(2);

    if (dfs_files_dont_exist(frameset))
        fors_flat_exit(NULL);


    cpl_msg_info(recipe, "Check input set-of-frames:");
    cpl_msg_indent_more();

    nbias = cpl_frameset_count_tags(frameset, master_bias_tag);
    if (nbias == 0) {
        cpl_msg_error(recipe, "Missing required input: %s", master_bias_tag);
        fors_flat_exit(NULL);
    }
    if (nbias > 1) {
        cpl_msg_error(recipe, "Too many in input (%d > 1): %s", 
                      nbias, master_bias_tag);
        fors_flat_exit(NULL);
    }

    nflat  = flat_mxu = cpl_frameset_count_tags(frameset, "SCREEN_FLAT_MXU");
    nflat += flat_mos = cpl_frameset_count_tags(frameset, "SCREEN_FLAT_MOS");
    nflat += flat_lss = cpl_frameset_count_tags(frameset, "SCREEN_FLAT_LSS");

    if (nflat == 0) {
        fors_flat_exit("Missing required input raw frames");
    }

    if (flat_mxu) {
        flat_tag = "SCREEN_FLAT_MXU";
        pro_tag = "MASTER_SCREEN_FLAT_MXU";
    }
    else if (flat_mos) {
        flat_tag = "SCREEN_FLAT_MOS";
        pro_tag = "MASTER_SCREEN_FLAT_MOS";
    }
    else if (flat_lss) {
        flat_tag = "SCREEN_FLAT_LSS";
        pro_tag = "MASTER_SCREEN_FLAT_LSS";
    }

    if (!dfs_equal_keyword(frameset, "ESO INS GRIS1 ID"))
        fors_flat_exit("Input frames are not from the same grism");

    if (!dfs_equal_keyword(frameset, "ESO INS FILT1 ID"))
        fors_flat_exit("Input frames are not from the same filter");

    if (!dfs_equal_keyword(frameset, "ESO DET CHIP1 ID")) 
        fors_flat_exit("Input frames are not from the same chip");

    header = dfs_load_header(frameset, flat_tag, 0);

    if (header == NULL) {
        cpl_msg_error(recipe, "Cannot load header of %s frame", flat_tag);
        fors_flat_exit(NULL);
    }

    instrume = (char *)cpl_propertylist_get_string(header, "INSTRUME");
    if (instrume == NULL) {
        cpl_msg_error(recipe, "Missing keyword INSTRUME in %s header", 
                      flat_tag);
        fors_flat_exit(NULL);
    }

    if (instrume[4] == '1')
        snprintf(version, 80, "%s/%s", "fors1", VERSION);
    if (instrume[4] == '2')
        snprintf(version, 80, "%s/%s", "fors2", VERSION);


    cpl_msg_indent_less();
    cpl_msg_info(recipe, "Load input frames:");
    cpl_msg_indent_more();

    master_flat = dfs_load_image(frameset, flat_tag, CPL_TYPE_FLOAT, 0, 0);

    if (master_flat == NULL)
        fors_flat_exit("Cannot load flat field");

    for (i = 1; i < nflat; i++) {
        flat = dfs_load_image(frameset, NULL, CPL_TYPE_FLOAT, 0, 0);
        if (flat) {
            cpl_image_add(master_flat, flat);
            cpl_image_delete(flat); flat = NULL;
        }
        else
            fors_flat_exit("Cannot load flat field");
    }

    master_bias = dfs_load_image(frameset, 
                                 master_bias_tag, CPL_TYPE_FLOAT, 0, 1);
    if (master_bias == NULL)
        fors_flat_exit("Cannot load master bias");

    cpl_msg_indent_less();
    cpl_msg_info(recipe, "Subtract the master bias from sum flat frame...");
    cpl_msg_indent_more();

    overscans = mos_load_overscans_vimos(header, 1);

    if (nflat > 1) {
        multi_bias = cpl_image_multiply_scalar_create(master_bias, nflat);
        dummy = mos_remove_bias(master_flat, multi_bias, overscans);
        cpl_image_delete(multi_bias);
    }
    else {
        dummy = mos_remove_bias(master_flat, master_bias, overscans);
    }

    cpl_table_delete(overscans); overscans = NULL;
    cpl_image_delete(master_bias); master_bias = NULL;
    cpl_image_delete(master_flat); master_flat = dummy;

    if (master_flat == NULL) {
        cpl_msg_error(recipe, "Cannot remove bias from sum flat frame");
        fors_flat_exit(NULL);
    }

    cpl_msg_indent_less();

    cpl_propertylist_update_int(header, "ESO PRO DATANCOM", nflat);

    if (dfs_save_image(frameset, master_flat, pro_tag,
                       header, parlist, recipe, version))
        fors_flat_exit(NULL);

    cpl_image_delete(master_flat); master_flat = NULL;
    cpl_propertylist_delete(header); header = NULL;

    return 0;
}
